import re
import types

from django import forms
from parsley.widgets import ParsleyChoiceFieldRendererMixin

FIELD_TYPES = [
    (forms.URLField, "url"),
    (forms.EmailField, "email"),
    (forms.IntegerField, "digits"),
    (forms.DecimalField, "number"),
    (forms.FloatField, "number"),
]


FIELD_ATTRS = [
    ("min_length", "minlength"),
    ("max_length", "maxlength"),
    ("min_value", "min"),
    ("max_value", "max"),
]


def update_widget_attrs(field, prefix='data'):
    attrs = field.widget.attrs
    if field.required:
        if isinstance(field.widget, forms.widgets.RadioSelect):
            try:
                # django >= 1.11
                original_create_option = field.widget.create_option

                def new_create_option(self, name, value, label, selected, index, subindex=None, attrs=None):
                    if index == len(self.choices) - 1:
                        attrs = attrs or {}
                        attrs["{prefix}-required".format(prefix=prefix)] = "true"

                    return original_create_option(name, value, label, selected, index, subindex, attrs)

                field.widget.create_option = types.MethodType(new_create_option, field.widget)
            except AttributeError:
                # django < 1.11

                # Use a mixin, to try and support non-standard renderers if possible
                class ParsleyChoiceFieldRenderer(ParsleyChoiceFieldRendererMixin, field.widget.renderer):
                    parsley_namespace = prefix
                field.widget.renderer = ParsleyChoiceFieldRenderer
        else:
            attrs["{prefix}-required".format(prefix=prefix)] = "true"
            error_message = field.error_messages.get('required', None)
            if error_message:
                attrs["{prefix}-required-message".format(prefix=prefix)] = error_message

    if isinstance(field, forms.RegexField):
        attrs.update({"{prefix}-pattern".format(prefix=prefix): field.regex.pattern})

        error_message = field.error_messages.get('invalid', None)
        if error_message:
            attrs["{prefix}-pattern-message".format(prefix=prefix)] = error_message

        if field.regex.flags & re.IGNORECASE:
            attrs.update({"{prefix}-regexp-flag".format(prefix=prefix): "i"})

    if isinstance(field, forms.MultiValueField):
        for subfield in field.fields:
            update_widget_attrs(subfield)

    # Set {prefix}-* attributes for parsley based on Django field attributes
    for attr, data_attr, in FIELD_ATTRS:
        if getattr(field, attr, None):
            attrs["{prefix}-{0}".format(data_attr, prefix=prefix)] = getattr(field, attr)

            error_message = field.error_messages.get(attr, None)
            if error_message:
                attrs["{prefix}-{0}-message".format(data_attr, prefix=prefix)] = error_message

    # Set {prefix}-type attribute based on Django field instance type
    for klass, field_type in FIELD_TYPES:
        if isinstance(field, klass):
            attrs["{prefix}-type".format(prefix=prefix)] = field_type

            error_message = field.error_messages.get('invalid', None)
            if error_message:
                attrs["{prefix}-type-message".format(field_type, prefix=prefix)] = error_message


def parsley_form(form):
    prefix = getattr(getattr(form, 'Meta', None), 'parsley_namespace', 'data-parsley')
    for _, field in form.fields.items():
        update_widget_attrs(field, prefix)
    extras = getattr(getattr(form, 'Meta', None), 'parsley_extras', {})
    for field_name, data in extras.items():
        for key, value in data.items():
            if field_name not in form.fields:
                continue
            attrs = form.fields[field_name].widget.attrs
            if key == 'equalto':
                # Use HTML id for {prefix}-equalto
                value = '#' + form[value].id_for_label
            if isinstance(value, bool):
                value = "true" if value else "false"
            attrs["{prefix}-%s".format(prefix=prefix) % key] = value
    return form


def parsleyfy(klass):
    """
    A decorator to add {prefix}-* attributes to your form.fields
    """

    old_init = klass.__init__

    def new_init(self, *args, **kwargs):
        old_init(self, *args, **kwargs)
        parsley_form(self)

    klass.__init__ = new_init

    return klass



class ParsleyChoiceFieldRendererMixin(object):
    def __init__(self, name, value, attrs, choices):
        self.name = name
        self.value = value
        self.attrs = attrs
        self._choices = choices

    @property
    def choices(self):
        def choices_iter():
            choices_list = self._choices
            for idx, choice in enumerate(choices_list):
                if idx == len(choices_list) - 1:
                    self.attrs["{prefix}-required".format(prefix=self.parsley_namespace)] = "true"
                yield choice
        return choices_iter()



from django import template, VERSION

from ..decorators import parsley_form

register = template.Library()

if VERSION[:2] >= (1, 9):
    # in django 1.9 and above, the simple_tag can do assignment
    tag_decorator = register.simple_tag
else:
    # django 1.8 and below needs the assignment_tag
    tag_decorator = register.assignment_tag


@tag_decorator()
def parsleyfy(form):
    return parsley_form(form)



import re

from django import forms
from parsley.decorators import parsleyfy
from .models import Student


class TextForm(forms.Form):
    name = forms.CharField(required=True,)
    university = forms.CharField(required=False)


@parsleyfy
class TextForm2(forms.Form):
    name = forms.CharField(required=True)
    university = forms.CharField(required=False)


@parsleyfy
class FieldTypeForm(forms.Form):
    name = forms.CharField(min_length=3, max_length=30)
    url = forms.URLField()
    url2 = forms.URLField(required=False)
    email = forms.EmailField()
    email2 = forms.EmailField(required=False)
    age = forms.IntegerField()
    income = forms.DecimalField()
    income2 = forms.FloatField()
    topnav = forms.RegexField(regex="#[A-Fa-f0-9]{6}")
    topnav2 = forms.RegexField(regex=re.compile("#[a-z]+", re.IGNORECASE))
    some_num = forms.IntegerField(min_value=10, max_value=100)
    amount = forms.DecimalField(max_digits=12, decimal_places=2, required=True, max_value=999999999999.99, min_value=-999999999999.99)


@parsleyfy
class ExtraDataForm(forms.Form):
    name = forms.CharField()
    email = forms.EmailField()
    email2 = forms.EmailField()
    hide_errors = forms.CharField()

    class Meta:
        parsley_extras = {
            "name": {
                "error-message": "Name invalid",
            },
            "email2": {
                "equalto": "email",
                "equalto-message": "Must match",
            },
            "hide_errors": {
                "show-errors": False,
            },
        }


class ExtraDataMissingFieldForm(ExtraDataForm):
    def __init__(self, *args, **kwargs):
        del self.base_fields['email2']
        super(ExtraDataMissingFieldForm, self).__init__(*args, **kwargs)


@parsleyfy
class FormWithWidgets(forms.Form):
    description = forms.CharField(widget=forms.TextInput)
    blurb = forms.CharField(widget=forms.TextInput(attrs={
        "class": "highlight"}))

@parsleyfy
class StudentModelForm(forms.ModelForm):
    class Meta:
        model = Student
        fields = '__all__'

    def __init__(self, *args, **kwargs):
        super(StudentModelForm, self).__init__(*args, **kwargs)

    def save(self, *args, **kwargs):
        if not self.instance.name:
            self.name = "Luke Skywalker"
        return super(StudentModelForm, self).save(*args, **kwargs)

@parsleyfy
class FormWithCustomInit(forms.Form):
    description = forms.CharField()

    def __init__(self, *args, **kwargs):
        super(FormWithCustomInit, self).__init__(*args, **kwargs)
        self.fields["description"].initial = "Hello"



class FormWithCleanField(forms.Form):
    description = forms.CharField(widget=forms.TextInput)

    def clean_description(self):
        raise forms.ValidationError("Error")


def get_state_choices():
    return [("NY", "NY"), ("OH", "OH")]

@parsleyfy
class FormWithCustomChoices(forms.Form):
    state = forms.ChoiceField(widget=forms.Select(choices=[]))
    def __init__(self, *args, **kwargs):
        super(FormWithCustomChoices, self).__init__(*args, **kwargs)
        self.fields['state'] = forms.ChoiceField(
            choices=get_state_choices())

@parsleyfy
class FormWithRadioSelect(forms.Form):
    state = forms.ChoiceField(choices=get_state_choices(), widget=forms.RadioSelect)


@parsleyfy
class FormWithRadioSelectNotRequired(forms.Form):
    state = forms.ChoiceField(choices=get_state_choices(), required=False, widget=forms.RadioSelect)


@parsleyfy
class FormWithMedia(forms.Form):
    name = forms.CharField(required=True)

    class Media:
        js = ("jquery.min.js",)
        css = {"all": ("jquery.css",)}


@parsleyfy
class FormWithoutMedia(forms.Form):
    name = forms.CharField(required=True)


class SSNWidget(forms.MultiWidget):
    def __init__(self, *args, **kwargs):
        kwargs['widgets'] = [
            forms.TextInput(),
            forms.TextInput(),
            forms.TextInput(),
            ]
        super(SSNWidget, self).__init__(*args, **kwargs)

    def decompress(self, value):
        return value.split('-') if value else [None, None, None]


class SSN(forms.MultiValueField):
    widget = SSNWidget
    def __init__(self, *args, **kwargs):
        fields = (
            forms.RegexField(r'^(\d)+$', min_length=3, max_length=3),
            forms.RegexField(r'^(\d)+$', min_length=3, max_length=3),
            forms.RegexField(r'^(\d)+$', min_length=4, max_length=4),
        )
        super(SSN, self).__init__(fields=fields, *args, **kwargs)


@parsleyfy
class MultiWidgetForm(forms.Form):
    ssn = SSN()


@parsleyfy
class CustomErrorMessageForm(forms.Form):
    name = forms.CharField(error_messages={"max_length": "Please only 30 characters"}, max_length=30, required=False)
    email = forms.EmailField(error_messages={"invalid": "Invalid email"}, required=False)
    favorite_color = forms.CharField(error_messages={"required": "Favorite color is required"})


@parsleyfy
class CustomPrefixForm(forms.Form):
    name = forms.CharField(required=True)

    class Meta:
        parsley_namespace = 'custom'



from parsley.decorators import parsleyfy


class ParsleyAdminMixin(object):

    def get_form(self, *args, **kwargs):
        form = super(ParsleyAdminMixin, self).get_form(*args, **kwargs)
        return parsleyfy(form)



import re
import six

from django import forms
from django.template import Context, Template
from django.test import TestCase
from django.utils.translation import ugettext_lazy as _

from parsley.decorators import parsleyfy, parsley_form

from .forms import (TextForm, TextForm2, FieldTypeForm, ExtraDataForm,
        ExtraDataMissingFieldForm, FormWithWidgets, StudentModelForm,
        FormWithCleanField, FormWithCustomInit, FormWithCustomChoices,
        MultiWidgetForm, CustomErrorMessageForm,
        CustomPrefixForm, FormWithRadioSelect, FormWithRadioSelectNotRequired)


class ParsleyTestCase(TestCase):

    def assertAttrsEqual(self, a, b):
        for k in a.keys():  # ignore unspecified keys
            if k in b:
                if six.PY3:
                    x, y = str(a[k]), str(b[k])
                else:
                    x, y = unicode(a[k]), unicode(b[k])
                self.assertEqual(x, y)


class CharFieldTest(ParsleyTestCase):
    def test_basic(self):
        """
        Tests that parsleyfy will add data-parsley-required for required fields,
        but not for required=False fields for CharFields
        """
        form = TextForm()
        self.assertEqual(form.fields["name"].widget.attrs, {})
        self.assertEqual(form.fields["university"].widget.attrs, {})
        form = parsley_form(TextForm())
        self.assertAttrsEqual(form.fields["name"].widget.attrs, {
            "data-parsley-required": "true",
            "data-parsley-required-message": _("This field is required.")
        })
        self.assertEqual(form.fields["university"].widget.attrs, {})


class CharFieldDecoratedTest(ParsleyTestCase):
    def test_decorated(self):
        "Tests that parsleyfy works as a class Decorator"
        form = TextForm2()
        self.assertAttrsEqual(form.fields["name"].widget.attrs, {
            "data-parsley-required": "true",
            "data-parsley-required-message": _("This field is required.")
        })
        self.assertEqual(form.fields["university"].widget.attrs, {})


class FieldTypeFormTest(ParsleyTestCase):
    def test_fields(self):
        "Tests that parsleyfy adds data-parsley-required for things other than CharField"
        form = FieldTypeForm()
        fields = form.fields
        self.assertEqual(fields["url"].widget.attrs["data-parsley-required"], "true")
        self.assertFalse("data-parsley-required" in fields["url2"].widget.attrs)
        self.assertEqual(fields["email"].widget.attrs["data-parsley-required"], "true")
        self.assertFalse("data-parsley-required" in fields["email2"].widget.attrs)


class DataTypeTest(ParsleyTestCase):
    def test_data_types(self):
        "Test that different field types get correct data-parsley-type"
        form = FieldTypeForm()
        fields = form.fields
        self.assertTrue("data-parsley-type" in fields["url"].widget.attrs)
        self.assertEqual(fields["url"].widget.attrs["data-parsley-type"], "url")
        self.assertTrue("data-parsley-type" in fields["email"].widget.attrs)
        self.assertEqual(fields["email"].widget.attrs["data-parsley-type"], "email")
        self.assertEqual(fields["age"].widget.attrs["data-parsley-type"], "digits")
        self.assertEqual(fields["income"].widget.attrs["data-parsley-type"], "number")
        self.assertEqual(fields["income2"].widget.attrs["data-parsley-type"], "number")
        self.assertEqual(fields["topnav"].widget.attrs["data-parsley-pattern"], "#[A-Fa-f0-9]{6}")
        self.assertNotIn("data-parsley-regexp-flag", fields["topnav"].widget.attrs)
        self.assertEqual(fields["topnav2"].widget.attrs["data-parsley-pattern"], "#[a-z]+")
        self.assertEqual(fields["topnav2"].widget.attrs["data-parsley-regexp-flag"], "i")


class LengthTest(ParsleyTestCase):
    def test_length(self):
        form = FieldTypeForm()
        fields = form.fields
        name_attrs = fields["name"].widget.attrs
        self.assertTrue("data-parsley-minlength" in name_attrs)
        self.assertEqual(name_attrs["data-parsley-minlength"], 3)
        self.assertEqual(name_attrs["data-parsley-maxlength"], 30)
        self.assertEqual(fields["amount"].widget.attrs["data-parsley-max"], 999999999999.99)
        self.assertEqual(fields["amount"].widget.attrs["data-parsley-min"], -999999999999.99)

class ValueTest(ParsleyTestCase):
    def test_value(self):
        form = FieldTypeForm()
        fields = form.fields
        num_attrs = fields['some_num'].widget.attrs
        self.assertTrue("data-parsley-min" in num_attrs, True)
        self.assertTrue("data-parsley-max" in num_attrs, True)
        self.assertEqual(num_attrs["data-parsley-min"], 10)
        self.assertEqual(num_attrs["data-parsley-max"], 100)


class FormWithWidgetsTest(ParsleyTestCase):
    def test_widgets(self):
        "Assert that @parsleyfy doesn't cloober existing attrs"
        form = FormWithWidgets()
        self.assertTrue(form.fields["description"].widget, forms.TextInput)
        self.assertEqual("highlight", form.fields["blurb"].widget.attrs["class"])


class TestMetadata(ParsleyTestCase):
    def test_docstring(self):
        form1 = TextForm()
        form2 = parsley_form(TextForm())
        self.assertEqual(form1.__doc__, form2.__doc__)

    def test_module(self):
        form1 = TextForm()
        form2 = parsley_form(TextForm())
        self.assertEqual(form1.__module__, form2.__module__)

    def test_name(self):
        form1 = TextForm()
        form2 = parsley_form(TextForm())
        self.assertEqual(form1.__class__.__name__, form2.__class__.__name__)


class TestModelForm(ParsleyTestCase):
    def test_model_form(self):
        form = StudentModelForm()
        fields = form.fields
        foo_attrs = fields["name"].widget.attrs
        self.assertEqual(foo_attrs["data-parsley-required"], "true")

    def test_model_form_save(self):
        form = StudentModelForm({"name": "Luke Skywalker"})
        form.save(commit=False)


class TestCustomInit(ParsleyTestCase):
    def test_custom_init(self):
        form = FormWithCustomInit()
        self.assertEqual(form.fields["description"].initial, "Hello")

    def test_custom_choices(self):
        form = FormWithCustomChoices()
        self.assertNotEqual(len(form.fields['state'].choices), 0)
        self.assertEqual(form.fields['state'].choices,
                    [("NY", "NY"), ("OH", "OH")])


class TestRadioSelect(ParsleyTestCase):
    def test_radio_select(self):
        form = FormWithRadioSelect()
        self.assertEqual(form.fields['state'].choices,
                    [("NY", "NY"), ("OH", "OH")])
        radio_select_html = form.fields['state'].widget.render("state", "NY")
        self.assertEqual(1, len(re.findall('data-parsley-required', radio_select_html)))

    def test_radio_select_not_required(self):
        form = FormWithRadioSelectNotRequired()
        self.assertEqual(form.fields['state'].choices,
                    [("NY", "NY"), ("OH", "OH")])
        radio_select_html = form.fields['state'].widget.render("state", "NY")
        self.assertEqual(0, len(re.findall('data-parsley-required', radio_select_html)))


class TestCleanFields(ParsleyTestCase):
    def test_clean(self):
        form = FormWithCleanField(data={"description": "foo"})
        self.assertEqual(form.is_bound, True)
        self.assertEqual(form.is_valid(), False)
        self.assertTrue(hasattr(form, "clean_description"))

    def test_clean_parslyfied(self):
        form = parsleyfy(FormWithCleanField)(data={"description": "foo"})
        self.assertEqual(form.is_bound, True)
        self.assertEqual(form.is_valid(), False)
        self.assertTrue(hasattr(form, "clean_description"))


class TestExtraAttributes(ParsleyTestCase):
    def test_equalto(self):
        form = ExtraDataForm()
        attrs = form.fields["email2"].widget.attrs
        self.assertAttrsEqual(attrs, {
            "data-parsley-type": "email",
            "data-parsley-required": "true",
            "data-parsley-equalto-message": "Must match",
            "data-parsley-equalto": "#id_email",
            "data-parsley-required-message": _("This field is required."),
        })

    def test_default_data(self):
        form = ExtraDataForm()
        attrs = form.fields["name"].widget.attrs
        self.assertAttrsEqual(attrs, {
            "data-parsley-required": "true",
            "data-parsley-error-message": "Name invalid",
            "data-parsley-required-message": _("This field is required.")
        })

    def test_boolean_values(self):
        form = ExtraDataForm()
        attrs = form.fields["hide_errors"].widget.attrs
        self.assertAttrsEqual(attrs, {
            "data-parsley-required": "true",
            "data-parsley-show-errors": "false",
            "data-parsley-required-message": _("This field is required.")
        })

    def test_missing_field(self):
        ExtraDataMissingFieldForm()  # No error should be raised


class TestMultiValueField(ParsleyTestCase):
    def test_parsley_attributes(self):
        form = MultiWidgetForm()
        fields = form.fields["ssn"].fields
        self.assertAttrsEqual(fields[0].widget.attrs, {
            "data-parsley-minlength": 3,
            "data-parsley-maxlength": 3,
            "maxlength": "3",
            "data-parsley-pattern": r'^(\d)+$',
        })
        self.assertAttrsEqual(fields[1].widget.attrs, {
            "data-parsley-minlength": 3,
            "data-parsley-maxlength": 3,
            "maxlength": "3",
            "data-parsley-pattern": r'^(\d)+$',
        })
        self.assertAttrsEqual(fields[2].widget.attrs, {
            "data-parsley-minlength": 4,
            "data-parsley-maxlength": 4,
            "maxlength": "4",
            "data-parsley-pattern": r'^(\d)+$',
        })


class TestCustomErrorMessages(TestCase):

    def test_new_message(self):
        form = CustomErrorMessageForm()
        attrs = form.fields['name'].widget.attrs
        self.assertEqual(attrs, {
            "maxlength": '30',
            "data-parsley-maxlength": 30,
            "data-parsley-maxlength-message": "Please only 30 characters"
        })

    def test_field_type_message(self):
        form = CustomErrorMessageForm()
        attrs = form.fields['email'].widget.attrs
        self.assertEqual(attrs, {
            "data-parsley-type": "email",
            "data-parsley-type-message": "Invalid email"
        })

    def test_override_default_message(self):
        form = CustomErrorMessageForm()
        attrs = form.fields['favorite_color'].widget.attrs
        self.assertEqual(attrs, {
            "data-parsley-required": "true",
            "data-parsley-required-message": "Favorite color is required"
        })


class TestCustomPrefix(TestCase):

    def test_default_prefix(self):
        form = TextForm2()
        attrs = form.fields['name'].widget.attrs
        self.assertTrue('data-parsley-required' in attrs)

    def test_custom_prefix(self):
        form = CustomPrefixForm()
        attrs = form.fields['name'].widget.attrs
        self.assertTrue('custom-required' in attrs)


class TemplateTagTest(ParsleyTestCase):
    def test_basic(self):
        """
        Tests that parsleyfy will work using the template tag
        """

        template = Template("{% load parsley %}")
        form = TextForm()
        context = Context({'form': form})
        template.render(context)

        self.assertEqual(context['form'].fields["name"].widget.attrs, {})
        self.assertEqual(context['form'].fields["university"].widget.attrs, {})

        template = Template("{% load parsley %}{% parsleyfy form as form %}")
        form = TextForm()
        context = Context({'form': form})
        template.render(context)

        self.assertAttrsEqual(context['form'].fields["name"].widget.attrs, {
            "data-parsley-required": "true",
            "data-parsley-required-message": _("This field is required.")
        })
        self.assertEqual(context['form'].fields["university"].widget.attrs, {})